package org.votech.plastic.managers;

import java.io.DataInputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URL;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.votech.plastic.incoming.handlers.ExtendableHandler;
import org.votech.plastic.incoming.handlers.StandardHandler;
import org.votech.plastic.incoming.messages.votable.VOTableLoader;
import org.votech.plastic.incoming.messages.votable.VotableLoadFromUrlMessageInvoker;
import org.votech.plastic.managers.AbstractObservableManager.ManagerObserver;
import org.votech.plastic.outgoing.messages.info.EchoMessage;
import org.votech.plastic.outgoing.messages.info.GetDescriptionMessage;



/**
 * This class is an example application that uses the {@link org.votech.plastic.managers.PlasticConnectionManager} class
 * to handle all things to do with Plastic.
 * It responds only to the core Plastic Messages to do with returning name, description etc. and the loadVOTable message.
 * When you run the main method it will attempt to connect the PlasticHub, and then execute a method
 * that lists the currently connected Plastic applications.  Finally it will send an ECHO message to each
 * application in turn.   The class implements two interfaces that allow it to register as a listener for Hub starting/stopping
 * and application registering/unregistering events.  To see a simpler example see {@link org.votech.plastic.managers.SimplestExampleApplication}.
 * @author jdt
 * @see #sendSomeMessages()
 * @see org.votech.plastic.managers.AbstractObservableManager.ManagerObserver
 * @see org.votech.plastic.managers.PlasticRegistryListener
 *
 */
public class ExampleApplication implements ManagerObserver, PlasticRegistryListener {

    private static final String NAME = "Example";
    private static final String DESCRIPTION = "An example Plastic app showing the use of the PlasticManager";
    private static final String IVORN = "ivo://roe.ac.uk/plastic/jdt/example";
    private static final String LOGOURL = "";
    private PlasticConnectionManager manager;
    private PlasticRegistry registry;

    public ExampleApplication() {
        initPlastic();
    }
    /**
     * Set up the {@link PlasticConnectionManager} that is responsible for housekeeping tasks
     * such as keeping track of whether the hub is up or down, registering us, and responding
     * to common informational messages.
     *
     */
    private void initPlastic() {
        //Create a message handler for the core informational messages
        StandardHandler handler = new StandardHandler(NAME, DESCRIPTION, IVORN, LOGOURL);
        //Create another message handler for the loadVOTable message.  This type of handler can 
        //handle any number of different messages.
        ExtendableHandler exhandler = new ExtendableHandler();
        exhandler.addMessageInvoker(new VotableLoadFromUrlMessageInvoker(new VOTableLoader() {
            //A VOTable loader is something capable of loading a votable!  In this case, our
            //anon inner class just has a dummy method.
            public void loadVOTable(URI sender, String id, URL url) throws Exception {
                System.out.println("Application "+sender+" has asked us to load a table from "+url+" and give it ID "+id);
            }
            
        }));
        //Handlers can be then chained together.  Some handlers "absorb" messages that they have dealt with, while
        //other types of handlers such as the LoggingHandler, deal with them and pass them on to the next handler.
        handler.appendHandler(exhandler);
        
        //Create the Plastic manager, tell it your name, the handler at the beginning of the chain, the fact
        //that we don't want to connect immediately, and that should we lose the hub we want to retry to
        //connect every 10 seconds.
        manager = new PlasticConnectionManagerImpl(NAME, handler, false, 10000 );
        
        //Register ourselves as being interested in the hub starting and stopping
        //See ManagerObserver interface
        manager.addObserver(this);
        
        //Instruct the manager to connect to the hub.
        manager.connect(); // Blocks waiting for connection.  See also manager.connectWhenReady()
        
        //While a PlasticConnectionManager keeps track of the hub, a PlasticRegistry keeps track of
        //other plastic applications.
        registry = new PlasticRegistry(manager);
        // Register ourselves as being interested in knowing when applications register and deregister
        registry.addListener(this);
        
    }

    
    /**
     * See who else is registered and get their metadata, if any.  
     * Send the ECHO message to each.
     * List all applications that understand the "getDescription" message.
     *
     */
    public void sendSomeMessages() {
        Collection apps =  registry.getPlasticApplications();
        Iterator it = apps.iterator();
        System.out.println("Registered applications-------");
        while (it.hasNext()) {
            PlasticApplication app = (PlasticApplication) it.next();
            System.out.println("Name: "+app.getName());
            System.out.println("Description: "+app.getDescription());
            System.out.println("Plastic version: "+app.getPlasticVersion());
            System.out.println("Icon URL: "+app.getIconUrl());
            System.out.println("IVORN: "+app.getIvorn());
            System.out.println("Plastic ID: "+app.getPlid());
            System.out.println("Supported messages:");
            List msgs = app.getSupportedMessageIds();
            for (int i=0;i<msgs.size();++i) {
                System.out.println("\t"+msgs.get(i));
            }
            
            //Send an EchoMessage to the application, and convert the returned object to the correct class.
            //If you prefer to work with the raw message IDs, there's a method for that too.
            final Object rawResponse = app.sendMessage(new EchoMessage("Hello World"));
            //Most Message objects have a static convertReturn method that knows how to convert the returned Object
            //to something more appropriate.
            String response = EchoMessage.convertReturn(rawResponse);
            
            System.out.println("Sent an echo message, application returned: "+response);
            System.out.println();
        }

        //Broadcast a message to everyone.
        System.out.println("Broadcasting an echo message to everyone");
        Map responses = registry.broadcast(new EchoMessage("Goodbye cruel world"));
        Iterator respondingappsIt = responses.keySet().iterator();
        while (respondingappsIt.hasNext()) {
            PlasticApplication rApp = (PlasticApplication) respondingappsIt.next();
            System.out.println("Application "+rApp.getName()+" responded "+responses.get(rApp));
            
        }
        
        //The registry can also be used to get all the applications that support a particular message, for example
        //the getDescription message.
        Collection appsWithDescription = registry.getPlasticApplications(new GetDescriptionMessage().getURI());
        Iterator it2 = appsWithDescription.iterator();
        System.out.println("The following applications support the getDescription message");
        while (it2.hasNext()) {
            PlasticApplication app = (PlasticApplication) it2.next();
            System.out.println(app.getName());
        }
    }

    
    /**
     * Called when the hub goes down
     * @see org.votech.plastic.managers.AbstractObservableManager.ManagerObserver
     */
    public void down() {
        System.out.println("The hub has gone down - exiting");
        System.exit(0);
    }
    /**
     * Called when the hub comes up
     * @see org.votech.plastic.managers.AbstractObservableManager.ManagerObserver
     */
    public void up() {
        System.out.println("The hub is up");
        System.out.println("This hub's ID is "+manager.getHub().getHubId());
        System.out.println("My ID is "+ manager.getId());
    }

    /**
     * Called when an application registers or unregisters.
     * @see PlasticRegistryListener
     */
    public void applicationsChanged() {
        System.out.println("An application has registered/deregistered");
        
    }

    /**
     * Run me.
     * @param args
     * @throws IOException
     */
    public static void main(String[] args) throws IOException {
        
        ExampleApplication app = new ExampleApplication();
        
        
        //Wait for the user to hit return
        final DataInputStream dataInputStream = new DataInputStream(System.in);
        while (true) {
            app.sendSomeMessages();
            System.out.println("Hit return to go again, type q to exit");
            String stop = dataInputStream.readLine();
            if ("q".equalsIgnoreCase(stop)) System.exit(0);
        }
    }
}
